/*
 * servoboard.cpp
 *
 * Created: 27.08.2012 11:03:38
 *  Author: Unknown
 */ 


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <math.h>
#include <util/delay.h>
#include "uart.h"
#include "components/wiring.h"
#include "components/Servo.h"

uint8_t unhex(char c) {
	return c >= 'A' ? c - 'A' + 10 : c - '0';
}

#define unhex2(H,L) ((unhex(H) << 4) | unhex(L))
#define fixbin(X) (X == 255 ? 255 : (X - (X >= 0x0A ? 1 : 0) - (X >= 0x0D ? 1 : 0)))

inline void setupWDT() {
	wdt_disable();
	wdt_enable(WDTO_2S);
	wdt_reset();
}

inline void setupCPU() {
	// external crystal used
	OSCCAL += 1; // needed to fix UART offset
}

static Servo servoMotor, servoLeft, servoRight;
static boolean uartConnected = false;
static boolean motorKilled = true;

/* setServos(0 = no motor - 255 = full thrust,  0 = no brake pulled, 255 = brake fully pulled) */
void setServos(uint8_t motor, uint8_t left, uint8_t right) {
	#define MOTOR0 1000
	if (motor == 0) {
		servoMotor.writeMicroseconds(MOTOR0); // hacked for carrier sense
		if (uartConnected && motorKilled) {
			motorKilled = false;
			PORTD |= _BV(PD2);
		}			
	}		
	else if (!motorKilled) {
		servoMotor.write(motor); 
	} else {
		servoMotor.writeMicroseconds(MOTOR0);
		PORTD ^= _BV(PD2);
	}		
	servoLeft.write(255 - left); // inverted
	servoRight.write(right); // inverted
}

/* Uart handling */
#define RX_SIZE 16
static char rxBuf[RX_SIZE];
static uint8_t rxLen = 0;

uint8_t handleCommand(char cmd[], uint8_t len) {
	// Commands:
	// b Set Servo -- 4 bytes [0xE_][Motor][Left][Right]
	//   Set Servo -- 10 bytes: [:][Motor][Left][Right][XOR-Checksum]
	//   Ping -- [:][P]
	//   Echo -- [:][E][...]
	if (len < 2) { return 0; }
	
	// Check if this is a valid command
	char type = cmd[0];
	if ((type >> 4) == 0x0E && len == 4) { // Binary set speed (0xE* w/ crc)
		uint8_t motor = cmd[1];
		uint8_t left = cmd[2];
		uint8_t right = cmd[3];

		// Use a 4 bit crc by only looking at the lower nibble
		uint8_t cksum4 = (motor >> 4) ^ motor ^ 
			             (left >> 4) ^ left ^
						 (right >> 4) ^ right;
	
		// Check the checksum
		if (((cksum4 ^ type) & 0x0F) == 0) {
			setServos(fixbin(motor), fixbin(left), fixbin(right));
			return 1; // keep alive
		}		
	}
	else if (type == ':') {
		// ASCII addressing mode
		if (cmd[1] == 'E') { // echo back the command
			for (uint8_t i = 2; i < len; i++)
			  uart_putc(cmd[i]);
			 uart_putc('\n');
		}
		else if (cmd[1] == 'P' && len == 2) {
			// Ping command
			uart_putstr_P(PSTR(":PING\n"));
		}
	}
	return 0;
}

uint8_t uartPoll() {
	char chr;
	while (uart_getc_nb(&chr)) {
		if (chr == '\n' || chr == '\r') {
			uint8_t ret = handleCommand(rxBuf, rxLen);
			rxLen = 0;
			return ret;
		} else if (rxLen < RX_SIZE)
			rxBuf[rxLen++] = chr;
	}	
	return 0; // no new msg
}

void onUartConnect() {
	PORTD |= _BV(PD2);
	uartConnected = true;
}
void onUartDisconnect() {
	// motor off, left: 33%, right: 15%  --> slow left turn
	setServos(0x00, 85, 25);
	PORTD &= ~_BV(PD2);
	motorKilled = true;
	uartConnected = false;
}

/* Main loop */
int main(void)
{
    MCUSR = 0;
	setupWDT();
	setupCPU();	
	_delay_ms(50); // wait for some stability
	
	// Init/test LED
	DDRD |= _BV(PD2);
	for (uint8_t i = 0; i < 4; i++) {
	  PORTD |= _BV(PD2);
	  _delay_ms(50);
	  PORTD &= ~_BV(PD2);
	  _delay_ms(50);
	}	  

	// Init UART
	uart_init();
	uart_putc('\n');
	uart_putstr_P(PSTR(":SERVO INIT v1\n"));
	
	// Init servos
	servoMotor.init(PD3, &PORTD, &DDRD, 1090, 1770);
	servoRight.init(PD4, &PORTD, &DDRD, 1082, 2115);
	servoLeft.init(PD5, &PORTD, &DDRD, 864, 1790);
	
	// Init with 0 power value in case servos have frequency auto-sense
	onUartDisconnect();
	initServos();
	wdt_reset();
	_delay_ms(250);

	// SERVO TEST
	//servoLeft.write(0);
	//while (1) {
		//for (uint16_t i = 700; i < 2000; i += 100) {
			//wdt_reset();
			//servoRight.writeMicroseconds(i);
			//servoLeft.writeMicroseconds(i);
			//uart_putc('.');
			//_delay_ms(80);
		//}
		//uart_putc('\n');
	//}
	
	// Main Loop -- go into high performance mode
	wdt_disable();
	wdt_enable(WDTO_60MS);
	#define TIMEOUT 1700 // roughly 1.5 sec
	uint16_t uartTimeout = 0;
	while (1) { 
		_delay_ms(0.5);
		wdt_reset(); 
		
		if (uartPoll()) {
			if (uartTimeout == 0)
				onUartConnect();
			uartTimeout = TIMEOUT;
		}
		else if (uartTimeout > 0) {
			if (--uartTimeout == 0) {
				onUartDisconnect();
			}
		}
	}
}